#include "pix_blobtracker.h"
#include <string.h>
#include <math.h>

#if 1
#define debug post
#else
#define debug
#endif

CPPEXTERN_NEW_WITH_ONE_ARG(pix_blobtracker,t_floatarg, A_DEFFLOAT) 

/*------------------------------------------------------------

 pix_blobtracker

 ------------------------------------------------------------*

/*------------------------------------------------------------

Constructor
initializes the pixBlocks and pixBlobs

------------------------------------------------------------*/
pix_blobtracker :: pix_blobtracker(t_floatarg f){
  m_blobNumber = (int)f;
  if(m_blobNumber < 1)m_blobNumber = 6;

  m_blobsize = 0.001;

  currentBlobs = new Blob[m_blobNumber];
  lastBlobs = new Blob[m_blobNumber];

  m_infoOut = outlet_new(this->x_obj, &s_list);
  m_countOut = outlet_new(this->x_obj, &s_float);
}

/*------------------------------------------------------------

Destructor

------------------------------------------------------------*/
pix_blobtracker :: ~pix_blobtracker(){
  outlet_free(m_infoOut);
  outlet_free(m_countOut);
  if(currentBlobs)delete[]currentBlobs;
}

/*------------------------------------------------------------
makeBlobs
calculates the Blobs, maximal x and y values are set
------------------------------------------------------------*/
void pix_blobtracker :: makeBlob(Blob *pb, imageStruct &img, int x, int y){
  if(!pb)return;

  if(x<0 || y<0)return;
  if(x>img.xsize || y>img.xsize)return;

  pb->area++;
  if(x<pb->xmin())pb->xmin(x);
  if(x>pb->xmax())pb->xmax(x);
  if(y<pb->ymin())pb->ymin(y);
  if(y>pb->ymax())pb->ymax(y);

  img.SetPixel(y,x,chGray,0);

  if(pb->area > 10000){return;}
  for(int i = -1; i<= 1; i++){
    for(int j = -1; j <= 1; j++){
      if (img.GetPixel(y+i, x+j, chGray) > 0)
	{
	  makeBlob(pb, img, x+j, y+i);
	}
    }
  }
}

/*------------------------------------------------------------
addToBlobArray
adds a detected Blob to the blob list
------------------------------------------------------------*/
void pix_blobtracker :: addToBlobArray(Blob *pb, int blobNumber){
  if (blobNumber >= m_blobNumber){
    // look whether we can replace a smaller blob
    float min = pb->area;
    int index=-1;
    int i = m_blobNumber;
    while(i--)

      if (currentBlobs[i].area < min){
	min = currentBlobs[i].area;
	index = i;
      }
    if (index!=-1)currentBlobs[index] = *pb;
  } else {
    currentBlobs[blobNumber] = *pb;
  }
}
/*------------------------------------------------------------
showBlobs
shows all blobs (rectangles)
------------------------------------------------------------*/
void pix_blobtracker :: showBlobs(imageStruct &image){

  /*for(int index = 0; index < m_blobNumber; index++){
    if(currentBlobs[index].valid){
      for(int y = (int)currentBlobs[index].ymin();
	  y <= (int)currentBlobs[index].ymax(); y++){
	image.SetPixel(y, (int)currentBlobs[index].xmin(),0,250);
	image.SetPixel(y, (int)currentBlobs[index].xmax(),0,250);
      }
      for(int x = (int)currentBlobs[index].xmin();
	  x <= (int)currentBlobs[index].xmax(); x++){
	image.SetPixel((int)currentBlobs[index].ymin(),x,0,250);
	image.SetPixel((int)currentBlobs[index].ymax(),x,0,250);
      }
    }
    }*/
}

/*------------------------------------------------------------
  
render

------------------------------------------------------------*/
void pix_blobtracker :: processGrayImage(imageStruct &image){
  int blobNumber = 0;
  int blobsize = (int)(m_blobsize * image.xsize * image.ysize);
  bool firstTime;


  for(int y = 0; y < image.ysize; y++){
    for(int x = 0; x < image.xsize; x++){
      if (image.GetPixel(y,x,0) > 0)
	{
	  Blob *blob = new Blob();
	  blob->xmin(image.xsize);
	  blob->ymin(image.ysize);

	  makeBlob(blob, image, x, y);
	  if(blob->area > blobsize){
	    blob->valid = true;
	    addToBlobArray(blob, blobNumber);
	    blobNumber++;
	  }
	  if(blob)delete blob;
	}
    }
  }


  if(blobNumber >= m_blobNumber)blobNumber = m_blobNumber-1;
  //showBlobs(image);
  //tracking
  if(numberOfOldBlobs() == 0)
    {
      //post("No Old blobs");
      copyBlobArrays(currentBlobs, lastBlobs);
      while(blobNumber--){
	lastBlobs[blobNumber].setPosition(blobNumber);
      }
      return;
    }
  trackBlobs(image, blobNumber);
  copyBlobArrays(currentBlobs, lastBlobs);

  outlet_float(m_countOut, (t_float)(blobNumber));

  float scaleX = 1./image.xsize;
  float scaleY = 1./image.ysize;

  blobNumber = m_blobNumber;
  while (blobNumber--){
    if(currentBlobs[blobNumber].valid){
      t_atom*ap=alist;
      SETFLOAT(ap, blobNumber);
      SETFLOAT(ap+1, currentBlobs[blobNumber].xmid()*scaleX);
      if (image.upsidedown)
	SETFLOAT(ap+2, 1.0-currentBlobs[blobNumber].ymid()*scaleY);
      else
	SETFLOAT(ap+2, currentBlobs[blobNumber].ymid()*scaleY);
      
      SETFLOAT(ap+3, currentBlobs[blobNumber].getPosition());
      //SETFLOAT(ap+3, (currentBlobs[blobNumber].area)*scaleX*scaleY);
      
      outlet_list(m_infoOut,gensym("list"), 4, alist);
      currentBlobs[blobNumber].valid = false;
    }
  }
}

/*------------------------------------------------------------
numberOfOldBlobs
counts the number of blobs in the lastBlobs -array
------------------------------------------------------------*/
int pix_blobtracker :: numberOfOldBlobs(){
  int index = m_blobNumber;
  int number = 0;
  while(index--){
    if(lastBlobs[index].valid){
      number++;
    }
  }
  //post("number: %d", number);
  return(number);
}
/*------------------------------------------------------------
copyBlobArrays
------------------------------------------------------------*/
void pix_blobtracker :: copyBlobArrays(Blob *source, Blob *b){
  int index = m_blobNumber;
  while(index--){
      b[index] = source[index];
  }
}
/*------------------------------------------------------------

trackBlobs

------------------------------------------------------------*/
void pix_blobtracker :: trackBlobs(imageStruct &image, int blobs){
  int indexLast, indexCurr;
  double help = image.xsize * image.xsize * image.xsize + image.ysize * image.ysize * image.ysize;
  double min = 0;
  int position = 0;
  int helpBlobs[m_blobNumber];

  for(indexLast = 0; indexLast < m_blobNumber; indexLast++){
    for(indexCurr = 0; indexCurr < m_blobNumber; indexCurr++){
      if((currentBlobs[indexCurr].valid) && (lastBlobs[indexLast].valid)){
	if(!currentBlobs[indexCurr].rightPosition){
	  min = lastBlobs[indexLast].distance2(currentBlobs[indexCurr]);
	  if(min < help){
	    help = min;
	    position = indexCurr;
	  }
	}
      }
    }
    if(help != image.xsize * image.xsize + image.ysize * image.ysize){
      currentBlobs[position].setPosition(lastBlobs[indexLast].getPosition());
      currentBlobs[position].rightPosition = true;
      help = image.xsize * image.xsize + image.ysize * image.ysize;
    }
  }
  for(indexCurr = 0; indexCurr < blobs; indexCurr++){
    if(!currentBlobs[indexCurr].rightPosition){
      currentBlobs[indexCurr].setPosition(blobs + indexCurr);
    }
    else{
      currentBlobs[indexCurr].rightPosition = false;
    }
  }
  post("End of tracking.");
}

/////////////////////////////////////////////////////////
// static member function
//
/////////////////////////////////////////////////////////
void pix_blobtracker :: obj_setupCallback(t_class *classPtr){
  class_addmethod(classPtr, (t_method)&pix_blobtracker :: blobSizeMessCallback,
		  gensym("blobSize"), A_FLOAT, A_NULL);
}

/*------------------------------------------------------------
blobSizeMessCallback
------------------------------------------------------------*/
void pix_blobtracker :: blobSizeMessCallback(void *data, t_floatarg blobSize){
  GetMyClass(data)->blobSizeMess((int) blobSize);
}
/*------------------------------------------------------------
blobSizeMess
------------------------------------------------------------*/
void pix_blobtracker :: blobSizeMess(int blobSize){
  if((blobSize < 0)||(blobSize > 100))
    {
      post("Out of range!");
      return;
    }
  m_blobsize = blobSize/100000.0;
  post("Blobsize set to: %f * imagesize.", m_blobsize);
}
