#ifndef INCLUDE_PIX_BLOBTRACKER_H_
#define INCLUDE_PIX_BLOBTRACKER_H_

#include "Base/GemPixObj.h"
#include <stdio.h>


class GEM_EXTERN Blob {
 public:

  Blob(){
    m_xmin = 0; m_xmax = 0.0;
    m_ymin = 0; m_ymax = 0.0;
    area = 0;
    position = 0;
    valid = false;
    rightPosition = false;
  }
    void setPosition(int p){position = p;}
    int getPosition(){return position;}
    double xmin(){return m_xmin;}
    double xmax(){return m_xmax;}
    double ymin(){return m_ymin;}
    double ymax(){return m_ymax;}

    double xmid(){return (m_xmin+m_xmax)*0.5;}
    double ymid(){return (m_ymin+m_ymax)*0.5;}
    // the squared diameter of the blob
    double diameter2(){return (m_xmax-m_xmin)*(m_xmax-m_xmin)+(m_ymax-m_ymin)*(m_ymax-m_ymin);}
    // the diamter
    double diameter(){return sqrt(diameter());}

    // the squared distance to another blob
    double distance2(Blob b){
      return (b.xmid()-xmid())*(b.xmid()-xmid())+(b.ymid()-ymid())*(b.ymid()-ymid());
    }
    // the distance to another blob
    double distance(Blob b){return sqrt(distance2(b));}

    void xmin(double x){m_xmin=x;}
    void xmax(double x){m_xmax=x;}
    void ymin(double y){m_ymin=y;}
    void ymax(double y){m_ymax=y;}

    int area;

    bool valid; // 0=invalid; 1=ok;
    bool rightPosition;
 private :
    double m_xmin;
    double m_xmax;
    double m_ymin;
    double m_ymax;
    int position;
};

class GEM_EXTERN pix_blobtracker : public GemPixObj
{
  CPPEXTERN_HEADER(pix_blobtracker, GemPixObj)
		
    public:
  
  //////////
  // Constructor
  pix_blobtracker(t_float f);
  t_outlet        *m_infoOut, *m_countOut;

 protected:  
  //Destructor
  ~pix_blobtracker();

  void processGrayImage(imageStruct &image);
  int m_blobNumber;
  Blob *currentBlobs;
  Blob *lastBlobs;
  void showBlobs(imageStruct &image);
  int numberOfOldBlobs();
  void addToBlobArray(Blob *pblob, int blobNumber);
  void makeBlob(Blob *pb, imageStruct &img, int x, int y);
  void copyBlobArrays(Blob *source, Blob *b);
  void trackBlobs(imageStruct &image, int blobs);
  // for the output
  t_atom alist[4];

  // the minimum size of a blob (relative to the image)
  float m_blobsize;
 private: 
  static void blobSizeMessCallback(void *data, t_floatarg blobSize);
  void blobSizeMess(int blobSize);
};


#endif 	// for header file
